(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Position 
imports 
  "~~/src/HOL/Library/Infinite_Set"
  "$AFP/Show/Show_Instances"
begin

datatype pos = Empty ("\<epsilon>") | PCons "nat" "pos" (infixr "<#" 70)

primrec append :: "pos \<Rightarrow> pos \<Rightarrow> pos" (infixr "<#>" 70) where
  append_Empty: "\<epsilon> <#> q = q" |
  append_PCons:  "(i <# p) <#> q = i <# (p <#> q)"

abbreviation singleton :: "nat \<Rightarrow> pos" ("<_>") where "<i> == i<#\<epsilon>"
abbreviation snoc :: "pos \<Rightarrow> nat \<Rightarrow> pos" (infixl "#>" 70) where "p#>i == p <#> <i>"

lemma not_PCons_self[simp]: "p \<noteq> i<#p" by (induct p) auto
lemmas not_PCons_self2[simp] = not_PCons_self[symmetric]
lemma not_snoc_self[simp]: "p \<noteq> p#>i" by (induct p) auto
lemmas not_snoc_self2[simp] = not_snoc_self[symmetric]
lemma impossible_PCons: "size p \<le> size q \<Longrightarrow> p = i<#q = False" by (induct p) auto
lemma size_0_conv[iff]: "(0 = size p) = (p = \<epsilon>)" by (cases p) auto
lemma size_greater_0_conf[iff]: "(0 < size p) = (p \<noteq> \<epsilon>)" by (cases p) auto 

lemma append_assoc[simp]: fixes p q r :: "pos" shows "(p <#> q) <#> r = p <#> (q <#> r)" by (induct p) auto
lemma append_Empty2[simp]: "p <#> \<epsilon> = p" by (induct p) auto
lemma size_append[iff]: fixes p :: "pos" shows "size(p <#> q) = size p + size q" by (induct p) auto

interpretation semigroup_append: semigroup_add "op <#> :: pos \<Rightarrow> pos \<Rightarrow> pos" by unfold_locales simp
interpretation monoid_append: monoid_add "op <#> :: pos \<Rightarrow> pos \<Rightarrow> pos" "\<epsilon>" by unfold_locales simp+

lemma append_is_Empty_conv[iff]: "(p <#> q = \<epsilon>) = (p = \<epsilon> \<and> q = \<epsilon>)" by (induct p) auto
lemma Empty_is_append_conv[iff]: "(\<epsilon> = p <#> q) = (p = \<epsilon> \<and> q = \<epsilon>)" by (induct p) auto
lemma append_self_conv[iff]: "(p <#> q = p) = (q = \<epsilon>)" by (induct p) auto
lemma self_append_conv[iff]: "(p = p <#> q) = (q = \<epsilon>)" by (induct p) auto

lemma append_eq_append_conv[simp,no_atp]:
 fixes p q r s :: "pos"
 assumes "size p = size q \<or> size r = size s"
 shows "(p <#> r = q <#> s) = (p = q \<and> r = s)"
using assms proof (induct p arbitrary: q)
 case Empty thus ?case by (cases q,simp,force)
next
 case (PCons i p) thus ?case by (cases q,force,simp)
qed

lemma same_append_eq[iff]: fixes p :: "pos" shows "(p <#> q = p <#> r) = (q = r)" by simp
lemma same_append_eq'[iff]: fixes p :: "pos" shows "(q <#> p = r <#> p) = (q = r)" by simp

primrec rev :: "pos \<Rightarrow> pos"
where "rev \<epsilon> = \<epsilon>"
    | "rev(i<#p) = rev p #> i"

lemma rev_append[simp]: fixes p :: "pos" shows "rev(p <#> q) = rev q <#> rev p" by (induct p) auto
lemma rev_rev_ident[simp]: fixes p :: "pos" shows "rev(rev p) = p" by (induct p) auto
lemma rev_size[simp]: fixes p :: "pos" shows "size(rev p) = size p" by (induct p) auto
lemma rev_inj: assumes "rev xs = rev ys" shows "xs = ys" using arg_cong[OF assms, of rev] by simp

lemma rev_induct[case_names Empty snoc]:
 fixes P :: "pos \<Rightarrow> bool" assumes A: "P \<epsilon>" and B: "\<And>i q. P q \<Longrightarrow> P(q#>i)" 
 shows "P p"
using assms by (simplesubst rev_rev_ident[symmetric],induct rule: pos.induct[where pos = "rev p"]) simp_all

lemma [simp]: "i<#p \<noteq> \<epsilon>" by simp
lemma [simp]: "\<epsilon> \<noteq> i<#p" by simp

lemma [simp]: "p#>i \<noteq> \<epsilon>" by simp
lemma [simp]: "\<epsilon> \<noteq> p#>i" by simp

instantiation pos :: order begin
 definition less_eq_pos :: "pos \<Rightarrow> pos \<Rightarrow> bool" where "less_eq_pos p q == (\<exists>r. (p <#> r) = q)"
 definition less_pos    :: "pos \<Rightarrow> pos \<Rightarrow> bool" where "less_pos p q    == (p \<le> q \<and> p \<noteq> q)"
 
 instance by (intro_classes, auto simp: less_eq_pos_def less_pos_def)
end

lemma Empty_least [intro!, simp]:
  "(\<epsilon> :: pos) \<le> p"
  by (auto simp: less_eq_pos_def)

lemma less_eq_simps[simp]: 
  "p \<le> p <#> q" 
  "(p <#> q1 \<le> p <#> q2) = (q1 \<le> q2)" 
  "(i <# q1 \<le> \<epsilon>) = False"
  "(i <# q1 \<le> j <# q2) = (i = j \<and> q1 \<le> q2)"
  "(p <#> q \<le> p) = (q = \<epsilon>)" unfolding less_eq_pos_def by auto

lemma [code]: "(\<epsilon> :: pos) \<le> p = True" 
  "(i <# q1 \<le> \<epsilon>) = False"
  "(i <# q1 \<le> j <# q2) = (i = j \<and> q1 \<le> q2)" by auto

lemma less_simps[simp]: 
  "(p < p <#> q) = (q \<noteq> \<epsilon>)" 
  "(p <#> q1 < p <#> q2) = (q1 < q2)" 
  "(p < \<epsilon>) = False"
  "(i <# q1 < j <# q2) = (i = j \<and> q1 < q2)"
  "(p <#> q < p) = False" unfolding less_pos_def less_eq_pos_def by auto

lemma prefix_smaller[simp]: fixes p :: "pos" assumes "p < q" shows "size p < size q"
using assms unfolding less_pos_def less_eq_pos_def by auto

lemma size_induct:
 fixes p :: "pos"
 assumes "\<And>p. (\<forall>q. size q < size p \<longrightarrow> P q) \<Longrightarrow> P p" shows "P p"
using assms by (rule measure_induct[of "size"]) iprover

lemma prefix_induct:
 fixes p :: "pos"
 assumes "\<And>p. (\<forall>q<p. P q) \<Longrightarrow> P p" shows "P p"
using assms by (rule measure_induct[of "size"]) auto

interpretation pos: semigroup_mult "op <#> :: pos \<Rightarrow> pos \<Rightarrow> pos" by (unfold_locales) simp

interpretation pos: monoid_mult "\<epsilon> :: pos" "op <#> :: pos \<Rightarrow> pos \<Rightarrow> pos" by unfold_locales simp_all

instantiation pos :: one begin
  definition one_pos_def[simp]: "1 = \<epsilon>"
  instance by (intro_classes)
end

instantiation pos :: times begin
  definition times_pos_def[simp]: "times p q = p <#> q"
  instance by (intro_classes)
end

instantiation pos :: power begin
  instance by (intro_classes)
end

instantiation pos :: semigroup_mult begin
  instance proof
    fix a b c::pos show "a * b * c = a * (b * c)" by simp
  qed
end

lemma power_append_distr: "p^(m+n) = p^(m::nat) <#> p^n" by (induct m) auto

lemma power_pos_Suc: "p^(Suc n) = p^n <#> p"
proof -
  have "p ^ Suc n = p ^ (n + Suc 0)" by simp
  also have "... = p ^ n <#> p" unfolding power_append_distr by auto 
  finally show ?thesis .
qed

lemma power_subtract_less_eq: "(p::pos)^((n::nat) - m) \<le> p^n"
proof(cases "m \<ge> n",simp)
 case False
  hence "(n -m) + m = n" by auto
  thus ?thesis unfolding less_eq_pos_def using power_append_distr by metis
qed


lemma power_size: fixes p :: "pos" shows "size (p^n) = size p * n" 
 by (induct n, simp, auto)

primrec pos_prefix :: "pos \<Rightarrow> pos \<Rightarrow> pos option"
where pre_Empty: "pos_prefix \<epsilon> q = Some q"
    | pre_PCons: "pos_prefix (i <# p) q = 
     (case q of \<epsilon> \<Rightarrow> None | j <# q' \<Rightarrow> if i=j then pos_prefix p q' else None)" 

lemma prefix_sound:
 "pos_prefix p q = Some r \<Longrightarrow> p <#> r = q"
proof (induct p arbitrary: q,simp)
 case (PCons i p) note IH = this
  thus ?case
   proof(cases q,simp)
    case (PCons j q')
     with IH have pq':"pos_prefix p q' = Some r" using pre_PCons option.distinct(1) by (cases "i = j",simp,force)
     have "i = j" using PCons option.distinct(1) pre_PCons IH by (cases "i = j",simp,force) 
     with pq' IH PCons show "(i <# p) <#> r = q" by auto
  qed
qed

lemma prefix_exists:
  assumes "p \<le> q"
  shows "\<exists> q'. p <#> q' = q \<and> pos_prefix p q = Some q'"
using assms
proof (induct p arbitrary: q)
  case (PCons i p')
  from PCons have "\<exists> q'. q = i <# q'" by (auto simp: less_eq_pos_def)
  then obtain q' where q_is: "q = i <# q'" and leq': "p' \<le> q'" using PCons by auto
  then have res: "pos_prefix (i <# p') q = pos_prefix p' q'" by (auto simp: pos_prefix_def)
  then have "\<exists> r. p' <#> r = q' \<and> pos_prefix p' q' = Some r" using leq' and PCons by auto
  then obtain r where "p' <#> r = q'" and "pos_prefix p' q' = Some r" by auto
  then show ?case using q_is and res by auto 
qed simp

fun pos_suffix :: "pos \<Rightarrow> pos \<Rightarrow> pos option"
where "pos_suffix p q = (case pos_prefix (rev p) (rev q) of
 None \<Rightarrow> None | Some q' \<Rightarrow> Some (rev q'))" 

lemma suffix_sound: "pos_suffix p q = Some r \<Longrightarrow> r <#> p = q" (is "?l \<Longrightarrow> ?r")
proof-
 assume ?l
 hence "pos_prefix (rev p) (rev q) = Some (rev r)" using pos_suffix_def option.distinct(1) 
  by (cases "pos_prefix (rev p) (rev q)",simp,auto)
 thus ?thesis using prefix_sound rev_append rev_inj by auto
qed

lemma suffix_complete: 
 "\<exists>r. r <#> p = q \<Longrightarrow> \<exists>r. r <#> p = q \<and> pos_suffix p q = Some r" (is "?ls \<Longrightarrow> ?rs")
proof-
 assume ?ls
 hence "rev p \<le> rev q" using rev_append less_eq_pos_def by auto
 hence "\<exists> q'. (rev p) <#> q' = (rev q) \<and> pos_prefix (rev p) (rev q) = Some q'" using prefix_exists by auto
 then obtain r where r:"(rev p) <#> r = (rev q)" and "pos_prefix (rev p) (rev q) = Some r" by auto
 hence "pos_suffix p q = Some (rev r)" and "rev ((rev r) <#> p) = rev q" using rev_rev_ident rev_append by auto
 thus ?rs using rev_inj by auto
qed

text \<open>Hide unqualified names of constants that are already used elsewhere.\<close>
hide_const (open)  append rev

lemma pos_exhaust_rev[case_names empty snoc]:
  assumes "p = \<epsilon> \<Longrightarrow> P" and "\<And>\<pi> i. p = \<pi>#>i \<Longrightarrow> P" shows "P"
using assms by (induct p rule:rev_induct) auto

lemma epsilon_power_epsilon[simp]: "\<epsilon> ^ n = \<epsilon>" by (induct n, auto)

fun parallel_pos :: "pos \<Rightarrow> pos \<Rightarrow> bool" (infixr "\<bottom>" 64)
  where "\<epsilon> \<bottom> _ = False"
      | "_ \<bottom> \<epsilon> = False"
      | "i <# p \<bottom> j <# q = (i \<noteq> j \<or> p \<bottom> q)"

lemma parallel_pos: "p \<bottom> q = (\<not> p \<le> q \<and> \<not> q \<le> p)"
proof (induct p arbitrary: q)
  case Empty
  show ?case by simp
next
  case (PCons i p q) note IH = this
  show ?case
  proof (cases q)
    case Empty
    thus ?thesis by simp
  next
    case (PCons j q')
    show ?thesis 
    proof (cases "i = j")
      case False
      thus ?thesis unfolding PCons by simp
    next
      case True
      with IH[of q'] show ?thesis unfolding PCons by auto
    qed
  qed
qed

lemma parallel_pos_prefix: "p1 \<bottom> p2 \<Longrightarrow>
  \<exists> p i j q1 q2. p1 = p <#> i <# q1 \<and> p2 = p <#> j <# q2 \<and> i \<noteq> j"
proof (induct p1 arbitrary: p2)
  case Empty thus ?case by simp
next
  case (PCons i p1)
  from PCons(2) obtain j p2' where p2: "p2 = j <# p2'" by (cases p2, auto)
  with PCons(2) have "i \<noteq> j \<or> (i = j) \<and> p1 \<bottom> p2'" by simp
  thus ?case
  proof
    assume ij: "i \<noteq> j"
    show ?case
    proof (intro exI conjI)
      show "i <# p1 = \<epsilon> <#> i <# p1" by simp
    next
      show "p2 = \<epsilon> <#> j <# p2'" unfolding p2 by simp
    qed (rule ij)
  next
    assume "i = j \<and> p1 \<bottom> p2'" hence
      p12: "p1 \<bottom> p2'" and ij: "i = j" by auto
    from PCons(1)[OF p12] obtain p i' j' q1 q2 where p1: "p1 = p <#> i' <# q1"
      and p2': "p2' = p <#> j' <# q2" and ij': "i' \<noteq> j'" by blast
    show ?thesis
    proof (intro exI conjI)
      show "i <# p1 = (i <# p) <#> i' <# q1" unfolding p1 by simp
    next
      show "p2 = (i <# p) <#> j' <# q2" unfolding p2 p2' ij by simp
    qed (rule ij')
  qed
qed

lemma position_cases: "p \<le> q \<or> q < p \<or> p \<bottom> q"
proof (cases "p \<bottom> q")
  case True thus ?thesis by auto
next
  case False
  from this[unfolded parallel_pos] have "p \<le> q \<or> (\<not> p \<le> q \<and> q \<le> p)" by auto
  thus ?thesis
  proof
    assume "p \<le> q" thus ?thesis by auto
  next
    assume "\<not> p \<le> q \<and> q \<le> p" 
    hence "q < p" by auto
    thus ?thesis by auto
  qed
qed

lemma parallel_pos_sym: "p1 \<bottom> p2 \<Longrightarrow> p2 \<bottom> p1"
  unfolding parallel_pos by auto

lemma less_pos_def': "(p < q) = (\<exists> r. q = p <#> r \<and> r \<noteq> \<epsilon>)" (is "?l = ?r")
proof
  assume ?r then obtain r where id: "p <#> r = q" "r \<noteq> \<epsilon>" by blast
  show ?l unfolding less_pos_def less_eq_pos_def
    by (rule conjI[OF exI[of _ r]], unfold id[symmetric], insert id(2), auto)
next
  assume ?l
  then obtain r where id: "p <#> r = q" "p \<noteq> q" unfolding less_pos_def less_eq_pos_def by blast
  show ?r
    by (rule exI[of _ r], insert id(2), unfold id[symmetric], auto)
qed    

lemma pos_append_cases: 
  "p1 <#> p2 = q1 <#> q2 \<Longrightarrow> (\<exists> q3. p1 = q1 <#> q3 \<and> q2 = q3 <#> p2) \<or> 
      (\<exists> p3. q1 = p1 <#> p3 \<and> p2 = p3 <#> q2)"
proof (induct p1 arbitrary: q1)
  case Empty
  thus ?case by auto
next
  case (PCons i p1' q1) note IH = this
  show ?case 
  proof (cases q1)
    case Empty
    thus ?thesis using IH(2) by auto
  next
    case (PCons j q1')
    with IH(2) have id: "p1' <#> p2 = q1' <#> q2" and ij: "i = j" by auto
    from IH(1)[OF id]
    show ?thesis unfolding PCons ij by auto
  qed
qed

lemma pos_less_eq_append_not_parallel:
 assumes "q \<le> p <#> q'"
 shows "\<not> (q \<bottom> p)"
proof-
  from assms obtain r where "q <#> r = p <#> q'" unfolding less_eq_pos_def ..  
  hence dec:"(\<exists> q3. q = p <#> q3 \<and> q' = q3 <#> r) \<or> 
   (\<exists> p3. p = q <#> p3 \<and> r = p3 <#> q')" (is "?a \<or> ?b") by (rule pos_append_cases)
  hence "p \<le> q \<or> q \<le> p" unfolding less_eq_pos_def by blast
  thus ?thesis unfolding parallel_pos by auto
qed

lemma less_pos_power_split: "q < p^m \<Longrightarrow> \<exists> p' k. q = p^k <#> p' \<and> p' < p \<and> k < m"
proof (induct m arbitrary: q)
  case 0
  thus ?case by auto
next
  case (Suc n q)
  show ?case
  proof (cases "q < p")
    case True
    show ?thesis
      by (rule exI[of _ q], rule exI[of _ 0], insert True, auto)
  next
    case False
    from Suc(2) obtain r where pn: "p <#> p^n = q <#> r" unfolding less_pos_def' by auto
    from pos_append_cases[OF this]
    have "\<exists> r. q = p <#> r"
    proof
      assume "\<exists> s. p = q <#> s \<and> r = s <#> p ^ n"
      then obtain s where p: "p = q <#> s" by auto
      with False show ?thesis by auto
    qed auto
    then obtain r where q: "q = p <#> r" by auto
    with Suc(2) have "r < p ^ n" by simp
    from Suc(1)[OF this] obtain p' k where r: "r = p ^ k <#> p'" "p' < p" "k < n" by auto
    show ?thesis unfolding q
      by (rule exI[of _ p'], rule exI[of _ "Suc k"], insert r, auto)
  qed
qed

instantiation pos :: "show"
begin

fun shows_pos :: "pos \<Rightarrow> shows"
where
  "shows_pos (i <# \<epsilon>) = shows (Suc i)" |
  "shows_pos (i <# ps) = (shows (Suc i) +@+ shows_pos ps)" |
  "shows_pos \<epsilon> = shows_string ''epsilon''"

definition "shows_prec (d::nat) (p::pos) = shows_pos p"
definition "shows_list_pos (ps :: pos list) = showsp_list shows_prec 0 ps"

lemma shows_prec_pos_append [show_law_simps]:
  "shows_prec d (p::pos) (r @ s) = shows_prec d p r @ s"
proof (induct p arbitrary: r s)
    case (PCons i ps)
    then show ?case by (cases ps) (simp_all add: shows_prec_pos_def show_law_simps)
qed (simp add: shows_prec_pos_def show_law_simps)

instance by (default) (simp_all add: show_law_simps shows_list_pos_def)

end

fun proper_prefix_list :: "pos \<Rightarrow> pos list"
where
  "proper_prefix_list \<epsilon> = []" |
  "proper_prefix_list (PCons i p) = \<epsilon> # map (PCons i) (proper_prefix_list p)"

lemma proper_prefix_list[simp]: "set (proper_prefix_list p) = {q. q < p}"
proof (induction p)
  case Empty
  show ?case by auto
next
  case (PCons i p)
  note IH = this
  show ?case (is "?l = ?r")
  proof (rule set_eqI)
    fix q
    show "q \<in> ?l = (q \<in> ?r)"
    proof (cases q)
      case Empty
      have less: "\<epsilon> < i <# p" unfolding less_pos_def by auto
      show ?thesis unfolding Empty using less by auto
    next
      case (PCons j q')
      show ?thesis unfolding PCons by (auto simp: IH)
    qed
  qed
qed

definition prefix_list :: "pos \<Rightarrow> pos list" where
  "prefix_list p \<equiv> p # proper_prefix_list p"

lemma prefix_list[simp]: "set (prefix_list p) = { q. q \<le> p}"
  unfolding prefix_list_def by auto

definition bounded_postfixes :: "pos \<Rightarrow> pos list \<Rightarrow> pos list" where
  "bounded_postfixes p ps \<equiv> map the [opt\<leftarrow>map (pos_prefix p) ps . opt \<noteq> None]"

lemma bounded_postfixes[simp]: "set (bounded_postfixes p ps) = { r. p <#> r \<in> set ps}" (is "?l = ?r")
proof -
  note d = bounded_postfixes_def
  {
    fix r
    assume "r \<in> ?l"
    from this[unfolded d] obtain q res where q: "q \<in> set ps" and pre: "pos_prefix p q = res" and 
      res: "res \<noteq> None" and r: "r = the res" by auto
    from res r have res: "res = Some r" by (cases res, auto)
    from prefix_sound[OF pre[unfolded res]] q have "r \<in> ?r" by auto
  }
  moreover
  {
    fix r
    assume "r \<in> ?r"
    then obtain q where q: "q \<in> set ps" and p: "p <#> r = q" by auto
    hence "p \<le> q" by auto
    from prefix_exists[OF this] obtain q' where "p <#> q' = q \<and> pos_prefix p q = Some q'" by auto
    with p have "pos_prefix p q = Some r" by auto
    with \<open>q \<in> set ps\<close> have "r \<in> ?l" unfolding d by force
  }
  ultimately show ?thesis by blast
qed

definition left_of_pos :: "pos \<Rightarrow> pos \<Rightarrow> bool" 
 where "left_of_pos p q == (\<exists>r i j. r #> i \<le> p \<and> r #> j \<le> q \<and> i < j)"

lemma "\<not>(left_of_pos \<epsilon> p)" unfolding left_of_pos_def
 using Empty_is_append_conv by (metis Empty_least eq_iff pos.distinct(1))

lemma left_of_pos_append: "left_of_pos p q \<Longrightarrow> left_of_pos (p <#> p') (q <#> q')"
 unfolding left_of_pos_def by (metis dual_order.trans less_eq_simps(1))

lemma append_left_of_pos: "left_of_pos p q = left_of_pos (p' <#> p) (p' <#> q)"
 proof(rule iffI)
  assume "left_of_pos p q"
  thus "left_of_pos (p' <#> p) (p' <#> q)" unfolding left_of_pos_def by (metis less_eq_simps(2) semigroup_append.add_assoc)
 next
  assume "left_of_pos (p' <#> p) (p' <#> q)"
  thus "left_of_pos p q"
   proof(induct p' arbitrary: p q rule:rev_induct,simp)
    case (snoc a p')
    hence IH:"left_of_pos (p' <#> p) (p' <#> q) \<Longrightarrow> left_of_pos p q" and 
     left:"left_of_pos ((p' #> a) <#> p) ((p' #> a) <#> q)" by auto
    from left[unfolded left_of_pos_def] have "left_of_pos (p' <#> (a <# p)) (p' <#> (a <# q))"
     by (metis Position.append_assoc append_PCons monoid_append.add.left_neutral snoc.prems)
    with IH have "left_of_pos (a <# p) (a <# q)" unfolding left_of_pos_def by (metis left_of_pos_def snoc.hyps) 
    then obtain r i j r' r'' where x:"r <#> <i> <#> r' = a <# p" and y:"(r #> j) <#> r'' = a <# q" 
     and ij:"i < j" unfolding left_of_pos_def less_eq_pos_def by auto
    hence "r > \<epsilon>" unfolding less_pos_def'
     by (metis append_Empty append_PCons not_less_iff_gr_or_eq pos.inject)
    with x obtain rr where "r = a <# rr" using pos.exhaust[of r]
     by (metis less_eq_simps(1) less_eq_simps(4) less_simps(1) monoid_append.add.left_neutral)
    with x y have "rr <#> <i> <#> r' = p" and y:"(rr #> j) <#> r'' = q" by auto
    with ij show ?case unfolding left_of_pos_def less_eq_pos_def by auto
  qed
qed

lemma left_pos_parallel: "left_of_pos p q \<Longrightarrow> q \<bottom> p" unfolding left_of_pos_def
 proof-
  assume "\<exists>r i j. r #> i \<le> p \<and> r #> j \<le> q \<and> i < j"
  then obtain r i j where rp:"r #> i \<le> p" and rq:"r #> j \<le> q" and ij:"i < j" by auto
  from rp obtain p' where rp:"p = r <#> i <#  p'" unfolding less_eq_pos_def by auto
  from rq obtain q' where rq:"q = (r <#> (j <# q'))" unfolding less_eq_pos_def by auto
  from rp rq ij have pq:"\<not> p \<le> q" by force
  from rp rq ij have "\<not> q \<le> p" by force 
  with pq show ?thesis using parallel_pos by auto
 qed


lemma left_of_append_cases:" left_of_pos (p0 <#> p1) q \<Longrightarrow> p0 < q \<or> left_of_pos p0 q"
 proof-
  assume "left_of_pos (p0 <#> p1) q"
  then obtain r i j where rp:"r #> i \<le> (p0 <#> p1)" and rq:"r #> j \<le> q" and ij:"i < j" 
   unfolding left_of_pos_def by auto
  show ?thesis proof(cases "p0 \<le> r")
    case True
      with rq have "p0 < q" by (metis dual_order.strict_trans1 le_less_trans less_simps(1) not_PCons_self)
      thus ?thesis by auto
    next
    case False
      hence aux:"\<not> (\<exists> r'. p0 <#> r' = r)" unfolding less_eq_pos_def by auto
      from rp have par:"\<not> (r #> i \<bottom> p0)" using pos_less_eq_append_not_parallel by auto
      from aux have a:"\<not> (p0 \<le> r)" unfolding less_eq_pos_def by auto
      from rp have "\<not> (p0 \<bottom> r)" by (metis less_eq_simps(1) order.trans parallel_pos_sym pos_less_eq_append_not_parallel)
      with a have "p0 > r" using position_cases by auto
      then obtain oo where p0:"p0 = r <#> oo" and "oo > \<epsilon>" unfolding less_pos_def less_eq_pos_def by auto 
      have "\<not> (r #> i > p0)" unfolding less_pos_def less_eq_pos_def 
        by (metis \<open>p0 = r <#> oo\<close> a less_eq_simps(1) less_eq_simps(2) less_eq_simps(3) less_eq_simps(4) less_eq_simps(5) pos.exhaust)
      with par have "p0 \<ge> r #> i" using position_cases by auto
      with ij this[unfolded less_eq_pos_def] have "left_of_pos p0 q" unfolding left_of_pos_def using rq by auto
      thus ?thesis by auto
  qed
 qed

lemma append_left_of_cases: assumes left:"left_of_pos q (p0 <#> p1)" shows "p0 < q \<or> left_of_pos q p0"
 proof-
  from left obtain r i j where rp:"r #> i \<le> q" and rq:"r #> j \<le> (p0 <#> p1)" and ij:"i < j"
   unfolding left_of_pos_def by auto
  show ?thesis proof(cases "p0 \<le> r")
    case True
      with rp have "p0 < q" unfolding less_pos_def using not_snoc_self2 by (metis eq_iff less_eq_simps(1) order.trans)
      thus ?thesis by auto
    next
    case False
      hence aux:"\<not> (\<exists> r'. p0 <#> r' = r)" unfolding less_eq_pos_def by auto
      from rp rq have par:"\<not> (r #> j \<bottom> p0)" using pos_less_eq_append_not_parallel by auto
      from aux have a:"\<not> (p0 \<le> r)" unfolding less_eq_pos_def by auto
      from rq have "\<not> (p0 \<bottom> r)" by (metis less_eq_simps(1) order.trans parallel_pos_sym pos_less_eq_append_not_parallel)
      with a have "p0 > r" using position_cases by auto
      then obtain oo where p0:"p0 = r <#> oo" and "oo > \<epsilon>" unfolding less_pos_def less_eq_pos_def by auto 
      have "\<not> (r #> j > p0)" unfolding less_pos_def less_eq_pos_def using p0 a pos.exhaust[of p0] 
        by (metis less_eq_simps(1) less_eq_simps(2) less_eq_simps(3) less_eq_simps(4) less_eq_simps(5) pos.exhaust)
      with par have "p0 \<ge> r #> j" using position_cases by auto
      with ij this[unfolded less_eq_pos_def] have "left_of_pos q p0" unfolding left_of_pos_def using rp by auto
      thus ?thesis by auto
  qed
 qed


lemma parallel_imp_right_or_left_of: assumes par:"p \<bottom> q" shows "left_of_pos p q \<or> left_of_pos q p"
proof-
 from parallel_pos_prefix[OF par] obtain r i j p' q' where "p = r <#> i<# p'" and "q = r <#> j <# q'" 
  and ij:"i \<noteq> j" by blast
 hence "r <#> <i> \<le> p \<and> r <#> <j> \<le> q" by simp
 thus ?thesis unfolding left_of_pos_def using ij less_linear by blast
qed

lemma left_of_imp_not_right_of: assumes l:"left_of_pos p q" shows "\<not> left_of_pos q p"
proof
  assume l':"left_of_pos q p"
  from l obtain r i j where "r #> i \<le> p" and ij:"i < j" and "r #> j \<le> q" unfolding left_of_pos_def by blast
  then obtain p0 q0 where p:"p = (r #> i) <#> p0" and q:"q = (r #> j) <#> q0" unfolding less_eq_pos_def by auto
  from l' obtain r' i' j' where "r' #> j' \<le> p" and ij':"i' < j'" and "r' #> i' \<le> q" unfolding left_of_pos_def by blast
  then obtain p0' q0' where p':"p = (r' #> j') <#> p0'" and q':"q = (r' #> i') <#> q0'" unfolding less_eq_pos_def by auto
  from p p' have p:"r <#> (i <# p0) = r' <#> (j' <# p0')"  by auto
  from q q' have q:"r <#> (j <# q0) = r' <#> (i' <# q0')"  by auto
  with p ij ij' have ne:"r \<noteq> r'" using same_append_eq[of r] by (metis less_imp_not_less pos.inject)
  have nlt:"\<not> r < r'" proof
   assume "r < r'"
   then obtain r2 where r1:"r' = r <#> r2" and r2:"r2 \<noteq> \<epsilon>" unfolding less_pos_def less_eq_pos_def by auto
   from p have p':"i <# p0 = r2 <#> j' <# p0'" unfolding r1 append_assoc using less_eq_simps(2) by auto
   from q have q':"j <# q0 = r2 <#> i' <# q0'" unfolding r1 append_assoc using less_eq_simps(2) by auto
   from r2 obtain k rr where r2:"r2 = k<# rr" by (cases r2, auto)
   from p' q' ij pos.inject show False unfolding r2 by simp
  qed
  have "\<not> r' < r" proof
   assume "r' < r"
   then obtain r2 where r1:"r = r' <#> r2" and r2:"r2 \<noteq> \<epsilon>" unfolding less_pos_def less_eq_pos_def by auto
   from p have p':"r2 <#> i <# p0 = j' <# p0'" unfolding r1 append_assoc using less_eq_simps(2) by auto
   from q have q':"r2 <#> j <# q0 = i' <# q0'" unfolding r1 append_assoc using less_eq_simps(2) by auto
   from r2 obtain k rr where r2:"r2 = k<# rr" by (cases r2, auto)
   from p' q' ij' pos.inject show False unfolding r2 by simp
  qed
  with  nlt ne position_cases[of r r', unfolded order_le_less[of r r']] have "r \<bottom> r'" by fast
  with p q show False by (metis less_eq_simps(1) pos_less_eq_append_not_parallel)
qed

primrec is_left_of :: "pos \<Rightarrow> pos \<Rightarrow> bool"
where left_Empty: "is_left_of \<epsilon> q = False"
    | left_PCons: "is_left_of (i <# p) q = (case q of 
        \<epsilon> \<Rightarrow> False 
      | j <# q' \<Rightarrow> if i < j then True else if i>j then False else is_left_of p q')" 

lemma is_left_of: "is_left_of p q = left_of_pos p q"
proof
 assume l:"is_left_of p q"
 thus "left_of_pos p q" 
 proof(induct p arbitrary:q)
  case Empty
   with left_Empty show ?case by auto
  next
  case (PCons i p) note IH = this
   assume l:"is_left_of (i <# p) q"
   show ?case proof(cases q)
    case Empty 
     with l show ?thesis unfolding left_PCons by auto
    next
    case (PCons j q') 
     show ?thesis proof(cases "\<not> (i < j)", cases "j < i")
      case True
       with l PCons show ?thesis unfolding left_PCons by auto
      next
       assume "\<not> j < i" and "\<not> i < j"
       hence ij:"i = j" by auto
       with PCons l have "is_left_of p q'" unfolding left_PCons by auto
       with IH have "left_of_pos p q'" by blast
       with ij show "left_of_pos (i <# p) q" unfolding PCons left_of_pos_def 
        by (metis append_PCons less_eq_simps(4))
      next
       assume ij:"\<not> \<not> (i < j)"
       hence "\<epsilon> #> i \<le> (i <# p) \<and> \<epsilon> #> j \<le> (j <# q')" unfolding less_eq_pos_def by auto
       with PCons ij show ?thesis unfolding left_of_pos_def by blast
     qed
   qed
 qed
 next
 assume l:"left_of_pos p q"
 from this[unfolded left_of_pos_def] obtain r i j where "r #>i \<le> p" and "r #> j \<le> q" 
  and ij:"i < j" by blast
 then obtain p' q' where "p = (r #>i) <#> p'" and "q = (r #>j) <#> q'" unfolding less_eq_pos_def 
  by auto
 thus "is_left_of p q" proof(induct r arbitrary: p q p' q')
  case Empty
   assume p:"p = (\<epsilon> #> i) <#> p'" and q:"q = (\<epsilon> #> j) <#> q'"
   with l[unfolded p q append_Empty] show ?case using left_PCons ij by force
  next
  case (PCons k r') note IH = this
   assume p:"p = ((k <# r') #> i) <#> p'" and q:"q = ((k <# r') #> j) <#> q'"
   from ij have "left_of_pos ((r' #> i) <#> p') (( r' #> j) <#> q')" unfolding left_of_pos_def
    by (metis less_eq_pos_def)
   with IH have "is_left_of ((r' #> i) <#> p') (( r' #> j) <#> q')" by auto
   thus "is_left_of p q" unfolding p q  using left_PCons by force
  qed
qed

abbreviation right_of_pos :: "pos \<Rightarrow> pos \<Rightarrow> bool"
where
  "right_of_pos p q \<equiv> left_of_pos q p"

primrec to_list :: "pos \<Rightarrow> nat list"
where
  "to_list \<epsilon> = []" |
  "to_list (i <# p) = i # (to_list p)"

lemma pos_prefix_same [simp]:
  "pos_prefix p p = Some \<epsilon>"
  by (induct p) simp_all

definition "pos_diff p q = the (pos_prefix q p)"

lemma prefix_pos_diff [simp]:
  assumes "p \<le> q"
  shows "p <#> pos_diff q p = q"
  using prefix_exists [OF assms] by (auto simp: pos_diff_def)

lemma pos_diff_Empty2 [simp]:
  "pos_diff p \<epsilon> = p"
  by (auto simp: pos_diff_def)

lemma inj_nat_to_pos: "inj (rec_nat \<epsilon> PCons)" (is "inj ?f")
  unfolding inj_on_def
proof (intro ballI impI)
  fix x y
  show "?f x = ?f y \<Longrightarrow> x = y"
  proof (induct x arbitrary: y)
    case 0
    thus ?case by (cases y, auto)
  next
    case (Suc x sy)
    then obtain y where sy: "sy = Suc y" by (cases sy, auto)
    from Suc(2)[unfolded sy] have id: "?f x = ?f y" by auto
    from Suc(1)[OF this] sy show ?case by simp
  qed
qed

lemma infinite_UNIV_pos[simp]: "infinite (UNIV :: pos set)"
proof
  assume "finite (UNIV :: pos set)"
  from finite_subset[OF _ this, of "range (rec_nat \<epsilon> PCons)"]
    range_inj_infinite[OF inj_nat_to_pos]
  show False by blast
qed

end
